/*
Programm zur Erzeugung von Step-Impulsen für 3 Achsen zum Positionieren der Fräse (ohne Rechner)
    Copyright 2016  Harald Sattler

    Dieses Programm ist freie Software. Sie können es unter den Bedingungen der GNU General Public License, wie von der
    Free Software Foundation veröffentlicht, weitergeben und/oder modifizieren, entweder gemäß Version 3 der Lizenz
    oder (nach Ihrer Option) jeder späteren Version.

    Die Veröffentlichung dieses Programms erfolgt in der Hoffnung, daß es Ihnen von Nutzen sein wird, aber OHNE IRGENDEINE GARANTIE,
    sogar ohne die implizite Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK.
    Details finden Sie in der GNU General Public License.

    Sie sollten ein Exemplar der GNU General Public License zusammen mit diesem Programm erhalten haben.
    Falls nicht, siehe <http://www.gnu.org/licenses/>.

Nützlich, um VOR Beginn eines Fräsvorgangs das Werkzeug zu wechseln,
den Hantierraum frei zu fahren (Spindel aus dem Weg) oder den Nullpunkt anzufahren.

SerialComCNC (SCC) bekommt von dem Verfahren nichts mit, es ist _immer_ hinterher der Nullpunkt in SCC zu setzen!

Achtung!
Serielle Schnittstelle ist auf 57600 Baud eingestellt. Beachten bei Verwendung der Menus (z.B. Shift-Ctrl-M in der Arduino-IDE)

Version 1, 30.12.2015, Harald Sattler
  Die Step- und Richtungs-Impulse werden nur erzeugt, wenn der Enable-Schalter eingeschaltet ist
  um zu verhindern, dass aus Versehen während eines Fräsvorgangs zusätzliche Schritte erzeugt werden.
  Das würde das Werkstück unweigerlich zerstören.

  Die Step-Impulse werden über Interrupts, die Dir-Signale per Programm erzeugt.
  Die beiden Daumen-Joysticks werden im Analog IRQ abgefragt.

  Einige Parameter für die Step-Erzeugung lassen sich per serielle Schnittstelle ändern und im EPROM des AtMega ablegen
  Die Parametrierung und die Abfrage der Werte erfolgt über USB mittels virtueller COM-SS
  Unterstützte Befehle: mit ?<RETURN> wird eine kurze Hilfe angezeigt

Version 2, 13.01.2016
- Die Enable-Signale für die Interfaceplatine müssen vom Arduino erzeugt werden, da offenbar die Verbindung der drei
  ENABLE-Eingänge an der Interface-Platine die Steuerung über den GRBL-Arduino verhindert -> anderes Layout des breakout board
- Die Richtungsinformation für pos. Joystickwerte wird im EEPROM gespeichert und über die serielle Com-SS parametrierbar gemacht

Version 3, 17.01.2016
- Umstellung auf Timer und Counter für die Step-Erzeugung. Timer1 (16bit) läuft mit 100kHz. Mit Zählern wird die Step-Frequenz erzeugt
- Die Ausgänge werden Tri-State geschaltet, wenn keine Step-Impulse erzeugt werden sollen -> GRBL-Eingänge aktiv

Version 4, 23.01.2016
- Umstellung auf Timer2 (8bit), die möglichen Zähler-Werte sind ohnehin kleiner als 255
  -> Vorteil: keine Vorkehrungen für atomare Zugriffe auf Timer-Variablen notwendig
- Werte aus einer Tabelle, exponentielle oder sigmoidale Funktion (jetzt exponentiell)
- ADC-Start in den Timer-IRQ verlegt -> keine Interferenzen (meistens), das klappt aber nur für eine Achse, Verwirbelungen bei 2..3 Achsen :(

Version 5, 25.01.2016
- Verwerfen des Analog-IRQ, Joysticks werden in der Loop ausgelesen
  -> keine Interferenzen mit Timer-IRQ
- Schleichgang für langsames Herantasten an Hindernisse, ausgelöst mit TASTER

Version 5.1, 25.01.2016
- Umstellung auf Timer1 (16bit) für bessere Auflösung der Stepabfolge und langsamere kleinste Verfahrgeschwindigkeit
    -> alle betroffenen Vars unsigned int statt byte
- Schleichgang wird daduch überflüssig -> TASTER disabled
- Menu: Defaults und Richtungsflags sind auskommentiert (Platzprobs bei global Vars)
- Defaults angepasst

Version 5.2, 27.01.2016
- Ersatz der beiden Speed-Tabellen durch Berechnung der Timer-Parameter zur Laufzeit (die Formel ist übersichtlich
    und die Berechnungszeit dürfte nicht stören (testen!!)

    mit speedX [1..(hugeness-1)]  (nicht ab 0 weil mit speedX == 0 keine Steps erzeugt werden)
    mit stepsize im setup() und nach Änderung von Startwert oder Endwert berechnet
    speedtempX = speedvalues[speedX];            wird zu...
      j = ((hugeness-1) - speedX);
      speedtempX = ((1/(hugeness - j)) * ((startwert + (j * stepsize)) + startwert));
      mit j substituiert...
      speedtempX = (((float)1/(hugeness-((hugeness-1) - speedX))) * (startwert + (((hugeness-1) - speedX) * stepsize)) + startwert);
- Durch Wegfall der Tabellen kann das Menu wieder vollständig angezeigt und erweitert werden

- Umstellung der Timer-Berechnung von float auf Festkomma (28.01.2016)
    speedtempX = (((10000 / (hugeness-((hugeness-1) - speedX))) / 10000) * (startwert + (((hugeness-1) - speedX) * stepsize)) + startwert);
    -> Durchlaufzeit der Berechnung je Achse von 9,2 ms auf 2,2 ms
- Anzahl diskreter Step-Intervalle (unterschiedliche Timer-Werte, = hugeness) einstellbar per Menu

- Version der Software in Hilfe-Seite anzeigen (03.02.2016)

Version 5.3, 25.02.2016
- Festkommazahlen-Mathematik ist (in meiner Implementierung) langsamer als Float, daher wieder Float
  (evtl. kann die Berechnung aufgeteilt werden, so dass die kleinen Werte über Festkomma berechnet werden, nur die Großen über Float)
- Ausgabe der Entfernung in mm auf ein Display über Seriell (vorerst mal auf SerialComInstruments)
- Reset der Zähler mit grüner Taste

04.03.2016
- Anzahl Werte (hugeness) auf unsigned int geändert um mit sehr großem Endwert sehr langsam fahren zu können und trotzdem gute Auflösung
  bei hohen Stepraten zu haben (hat nicht geklappt, daher wieder rückgängig gemacht)

03.05.2016
- Reset der Zähler auf Timer umgestellt

Version 5.4, 14.05.2016
- Hilfetext im ROM abgelegt (RAM einsparen)

29.05.2016
- Anzahl Werte (hugeness) und Zähler j für "$$" auf unsigned int geändert um mit sehr großem Endwert sehr langsam fahren zu können
  und trotzdem gute Auflösung bei hohen Stepraten zu haben
- Beschleunigungs- und Abbremsrampe bei X eingebaut (Prototyp, funktioniert nicht)

05.06.2016
- Beschleunigungs- und Abbremsrampe bei X, Y und Z eingebaut (funktioniert)
- Watchdog-IRQ für Serial-Problem bei SerialComInstruments (klappt nicht)

Version 5.5, 21.08.2016
- manuelles Verfahren zum Fräsen von geraden Kanten erfordert, dass nicht versehentlich die andere horizontale Achse bewegt wird.
  Vertikal kann weiterhin immer verfahren werden, die Blockade gilt nur für die horizontalen Achsen.
  Um das zu gewährleisten wird zu Beginn die grüne Taste am Controller gedrückt und dann die Achse per Joystick betätigt, die verwendet
  werden soll. Nach Erkennen der gewünschten Achse blinkt der grüne Taster und die andere horizontale Achse wird nicht mehr bewegt.
  Abschalten dieses Modus durch erneuten Druck auf den grünen Taster -> Blinken hört auf
  
27.06.2017  
- zusätzliches Flag um die Weg-Werte für SerialComInstruments abzuschalten
- Achtung! SCHALTER-Abfrage ist disabled - if(1==1)

mögliche Erweiterungen:
- Endschalter beim Verfahren abfragen
- Probe abfragen (eher unsinnig, das kann SerialComCNC besser)
- temporären Nullpunkt setzen und per Button wieder anfahren (Übernahme der Kontrolle auf Joysticks für Werkzeugwechsel)
    -> hier müssen Beschleunigungsrampen gefahren werden -> AccelStepper-Lib?
*/

#include <EEPROM.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
/* #include <AccelStepper.h>
// Define a stepper and the pins it will use
AccelStepper stepperX(1, 2, 4);
AccelStepper stepperY(1, 5, 7);
AccelStepper stepperZ(1, 8, 10);
 */
// defines
#define MajorVersion 5
#define MinorVersion 5

// Definition der Ausgänge (Arduino Nummerierung und AtMega Nummerierung)
#define dirX   4                        // Pin D4  = Port D4 = PortD Bit 4
#define stepX  2                        // Pin D2  = Port D2 = PortD Bit 2
#define dirY   7                        // Pin D7  = Port D7 = PortD Bit 7
#define stepY  5                        // Pin D5  = Port D5 = PortD Bit 5
#define dirZ  10                        // Pin D10 = Port B2 = PortB Bit 2
#define stepZ  8                        // Pin D8  = Port B0 = PortB Bit 0
#define LED   13                        // Pin D13 = Port B5 = PortB Bit 5
#define GREENLED 18                     // Pin D18/A4 ist Ausgang für Grüne LED im Taster

#define SCHALTER  19                    // Pin D19/A5 ist Eingang für den Enable-Schalter
#define TASTER  17                      // Pin D17/A3 ist Eingang für den Taster (bisher nicht verwendet)

#define EnableX 12                      // drei getrennte Enable-Ausgänge für das Interface
#define EnableY 11
#define EnableZ 9

#define MIDDLE_INTERVAL 40              // Totraum um die Mittelstellung der Potis


// ------------------------------------------------------------------------------------------------
// Hilfsroutinen Watchdog zur Überwachung der SerialEvent-Routine
//{ mit diesen kommentierten Klammern kann der Code in Notepad++ ausgeblendet werden - bessere Lesbarkeit des Code
/** The following code was taken from the avr/wdt.h file and altered such as to
 * 	use _BV(WDIE) rather than _BV(WDE)
 * 	(see table 9-1 in the Atmel mega88 manual, page 54)
 */
#define wdt_set(value)   \
    __asm__ __volatile__ (  \
        "in __tmp_reg__,__SREG__" "\n\t"    \
        "cli" "\n\t"    \
        "wdr" "\n\t"    \
        "sts %0,%1" "\n\t"  \
        "out __SREG__,__tmp_reg__" "\n\t"   \
        "sts %0,%2" \
        : /* no outputs */  \
        : "M" (_SFR_MEM_ADDR(_WD_CONTROL_REG)), \
        "r" (_BV(_WD_CHANGE_BIT) | _BV(WDE)), \
			   "r" ((uint8_t) ((value & 0x08 ? _WD_PS3_MASK : 0x00) | \
					 _BV(WDIE) | (value & 0x07)) ) \
        : "r0"  \
    )
 //wdt_set

#define wdt_off(x)   \
    __asm__ __volatile__ (  \
        "in __tmp_reg__,__SREG__" "\n\t"    \
        "cli" "\n\t"    \
        "wdr" "\n\t"    \
        "sts %0,%1" "\n\t"  \
        "sts %0,%2" \
        : /* no outputs */  \
        : "M" (_SFR_MEM_ADDR(_WD_CONTROL_REG)), \
        "r" (_BV(_WD_CHANGE_BIT) | _BV(WDE)), \
			   "r" ((uint8_t) (0x00)) \
        : "r0"  \
    )
 //wdt_off
//} mit diesen kommentierten Klammern kann der Code in Notepad++ ausgeblendet werden - bessere Lesbarkeit des Code

// globale Variablen
String inputString = "";                // Speicher für Eingabestrings auf der seriellen SS
boolean stringComplete = false;         // CR/LF empfangen

int directionX;                         // Festlegung der Bewegungsrichtung der Stepper
int directionY;
int directionZ;
int eeAdr = 0;                          // Startadresse im EEPROM für Parameter
int debugprint = 0;                     // vorbelegen mit "keine Debugausgaben auf der seriellen SS"

int speedX;                             // aktuell mit dem Stick eingestelle Speed
int speedY;
int speedZ;
unsigned long tasterTime;				// Timer für Verzögertes Auslösen des Weg-Reset (Koordinaten auf Null setzen)
byte tasterLatch = 0;
unsigned long ledblinker = 0;

volatile unsigned int joystickX = 512;	// Mittelwerte vorbelegen für ersten Durchlauf
volatile unsigned int joystickY = 512;
volatile unsigned int joystickZ = 512;
boolean lockX = false;                  // Flags für manuelles Verfahren, die jeweils andere Achse wird blockiert
boolean lockY = false;
boolean blinkGreen = false;             // manuelles Verfahren wird mit grüner LED blinkend signalisiert
boolean manuellMode = false;
byte wegvalues = 0;						// 0: keine Ausgaben fuer SerialComInstruments


volatile unsigned int speedtempX;       // vars werden in der IRQ-Routine verwendet -> volatile
volatile unsigned int speedtempY;
volatile unsigned int speedtempZ;
volatile unsigned int uscounterX = 0;
volatile unsigned int uscounterY = 0;
volatile unsigned int uscounterZ = 0;

byte i;
unsigned int startwert;                 // kleinster Wert für speedarray
unsigned int endwert;                   // größter Wert für speedarray
unsigned int hugeness;                  // Anzahl diskreter Step-Intervalle (unterschiedliche Timer-Werte)
unsigned int stepsize;                  //  = (endwert - startwert) / (hugeness-1);

unsigned int maxincspeedStep = 500;   	// maximale Beschleunigung der Steps
unsigned int maxdecspeedStep = 500;   	// maximale Abbremsung der Steps
unsigned int sollspeedX;
unsigned int sollspeedY;
unsigned int sollspeedZ;
unsigned int oldspeedX;
unsigned int oldspeedY;
unsigned int oldspeedZ;

volatile long wegX = 0L;                // Zähler für Steps um zum Ausgangspunkt des Wegs zurück zu fahren (per AccelStepper Lib)
volatile long wegY = 0L;
volatile long wegZ = 0L;
long wegXsave = 0L;                     // Merker für Differenz
long wegYsave = 0L;
long wegZsave = 0L;
float wegXmm = 0L;                  	// Weg in mm
float wegYmm = 0L;
float wegZmm = 0L;
char wegBuffer[20];
volatile signed char Xdir = 0;
volatile signed char Ydir = 0;
volatile signed char Zdir = 0;

// Hilfetexte im ROM abspeichern (und so wertvolles RAM sparen)
//{
const char hilfetext0[] PROGMEM = "\nHilfe";
const char hilfetext1[] PROGMEM = "Folgende Befehle werden verstanden (Gross/Kleinschreibung ist zu beachten!):";
const char hilfetext2[] PROGMEM = "? / help     -> diese Hilfe";
const char hilfetext3[] PROGMEM = "$$           -> listet Parameter";
const char hilfetext4[] PROGMEM = "startwert=xx -> neuer unterer Wert (schnelles Stepping) fuer SpeedTable (*)";
const char hilfetext5[] PROGMEM = "endwert=xxx  -> neuer oberer Wert (langsames Stepping) fuer SpeedTable (*)";
const char hilfetext6[] PROGMEM = "stepint=xxx  -> neue Anzahl diskreter StepIntervalle (*)";
const char hilfetext7[] PROGMEM = "dirX=x       -> Laufrichtung bei positiven Joystick X Werten. x=[-1|1] (*)";
const char hilfetext8[] PROGMEM = "dirY=x       -> Laufrichtung bei positiven Joystick Y Werten. x=[-1|1] (*)";
const char hilfetext9[]  PROGMEM = "dirZ=x       -> Laufrichtung bei positiven Joystick Z Werten. x=[-1|1] (*)";
const char hilfetext10[] PROGMEM = "save         -> Parameter mit(*) werden ins EEPROM gesichert";
const char hilfetext11[] PROGMEM = "recall       -> Parameter mit(*) werden aus dem EEPROM geladen";
const char hilfetext12[] PROGMEM = "defaults     -> Parameter werden mit Defaults vorbelegt";
const char hilfetext13[] PROGMEM = "start        -> Debugausgaben starten";
const char hilfetext14[] PROGMEM = "stop         -> Debugausgaben anhalten\n";

PGM_P const array[15] PROGMEM = {
hilfetext0, hilfetext1, hilfetext2, hilfetext3, hilfetext4, hilfetext5, hilfetext6, hilfetext7, hilfetext8, hilfetext9,
hilfetext10, hilfetext11, hilfetext12, hilfetext13, hilfetext14};

char HilfeBuf[80];   // temporärer Buffer für die einzelnen Hilfetext-Zeilen
PGM_P pHilfeZeiger;
//}

void setup() {
  for(i=0; i++; i<20) {wegBuffer[i] = 0;}
 // set pin modes
  pinMode(SCHALTER, INPUT_PULLUP);      // Eingang für Schalter (== Enable Joysticks)
                                        // ohne eingeschalteten Schalter (dessen LED leuchtet ROT) werden keine Steps erzeugt
  pinMode(TASTER, INPUT_PULLUP);

  pinMode(A0, INPUT);                   // Analogeingänge als Eingang definieren
  pinMode(A1, INPUT);
  pinMode(A2, INPUT);

  pinMode(EnableX,INPUT);               // Signal EnableX zum Interface
  pinMode(EnableY,INPUT);               // Signal EnableY zum Interface
  pinMode(EnableZ,INPUT);               // Signal EnableZ zum Interface

  pinMode(LED,OUTPUT);                  // Ausgang LED
  pinMode(GREENLED,OUTPUT);             // Ausgang für Anzeige von Step-Aktivität (grüne LED im Schalter)

  pinMode(dirX,OUTPUT);                 // Ausgang Dir X
  pinMode(dirY,OUTPUT);                 // Ausgang Dir Y
  pinMode(dirZ,OUTPUT);                 // Ausgang Dir Z
  pinMode(stepX,OUTPUT);                // Ausgang Step X
  pinMode(stepY,OUTPUT);                // Ausgang Step Y
  pinMode(stepZ,OUTPUT);                // Ausgang Step Z

  Serial.begin(57600);                  // fuer Debugging und Parametrierung
  inputString.reserve(20);              // 20 byte fuer inputString reservieren

  eeAdr = 0;                            // Variablen aus dem EEPROM vorbelegen
  startwert = eepromReadInt(eeAdr);
  eeAdr += 2;
  endwert = eepromReadInt(eeAdr);
  eeAdr += 2;
  directionX = eepromReadInt(eeAdr);
  eeAdr += 2;
  directionY = eepromReadInt(eeAdr);
  eeAdr += 2;
  directionZ = eepromReadInt(eeAdr);
  eeAdr += 2;
  hugeness = eepromReadInt(eeAdr);

  for(i=0; i++; i<20) {
    digitalWrite(LED,1);
    delay(200);
    digitalWrite(LED,0);
    delay(200);
    }

// setup arrays for stepsize
calcStepsize();

cli();                                      // stop interrupts
{//setup timer1 (16 bit) interrupt --------------------------------------------------------------------------------------------
  TCCR1A = 0;                               // Timer/Counter1 Control Register A, set entire TCCR1A register to 0
  TCCR1B = 0;                               // Timer/Counter1 Control Register B, set entire TCCR1B register to 0
  TCCR1B |= (1 << WGM12);                   // turn on CTC mode
  // ----------------------
  // TCCR1B |= (1 << CS12);                 // TMR Prescaler Select Bits  CS22   CS21   CS20              Timerrate
  // TCCR1B |= (1 << CS11);                    //                              1      1      1    -> 1024 ->  15,625 kHz
  TCCR1B |= (1 << CS10);                    //                              1      1      0    -> 256  ->  62,5 kHz
                                            //                              1      0      1    -> 128  ->  125 kHz
                                            //                              1      0      0    -> 64   ->  250 kHz
                                            //                              0      1      1    -> 32   ->  500 kHz
                                            //                              0      1      0    -> 8    ->  2 MHz
                                            //                              0      0      1    -> 1    ->  16 MHz
  // ----------------------
  TCCR1C = 0;                               // Timer/Counter1 Control Register C, set entire TCCR1C register to 0
  TCNT1  = 0;                               // initialize counter value to 0
  OCR1A = 159;                              // set compare match register for xx khz increments
  // OCR1B =  timer1MaxVal;                             // set compare match register to 200 for 10 khz increments
  TIMSK1 |= (1 << OCIE1A);                  // enable timer1 compare interrupt A (Y-Achse)
  // TIMSK1 |= (1 << OCIE1B);                  // enable timer1 compare interrupt B (Y-Achse)
}

// Watchdog einschalten um Hänger in der SerialEvent-Routine zu vermeiden
wdt_set(WDTO_2S);

sei();                                      // enable interrupts again
} // setup() end -----------------------------------------------------------------------------------------------------

ISR(TIMER1_COMPA_vect){                     //timer1 interrupt alle 10 µs, 100 kHz (prescaler 1, 159 TCC)
  // PORTB |= (1 << PB5);                      // LED-Pin setzen
      // X-Achse
      if(speedtempX != 0)
      {
        if(uscounterX++ > speedtempX)
        {
          PORTD |= (1 << PD2);              // Step-Pin setzen
          // wegX += Xdir;
        }
        if(uscounterX++ > speedtempX + 3)
        {
          PORTD &= ~(1 << PD2);             // Step-Pin löschen
          uscounterX = 0;
        }
      }
      // Y-Achse
      if(speedtempY != 0)
      {
        if(uscounterY++ > speedtempY)
        {
          PORTD |= (1 << PD5);              // Step-Pin setzen
          // wegY += Ydir;
        }
        if(uscounterY++ > speedtempY + 1)
        {
          PORTD &= ~(1 << PD5);             // Step-Pin löschen
          uscounterY = 0;
        }
      }
      // Z-Achse
      if(speedtempZ != 0)
      {
        if(uscounterZ++ > speedtempZ)
        {
          PORTB |= (1 << PB0);              // Step-Pin setzen
          // wegZ += Zdir;
        }
        if(uscounterZ++ > speedtempZ + 1)
        {
          PORTB &= ~(1 << PB0);             // Step-Pin löschen
          uscounterZ = 0;
        }
      }
  // PORTB &= ~(1 << PB5);                     // LED-Pin loeschen
}

/******************************************************
 * WDT_vect interrupt handler
 * This handles the software watchdog timer interrupt
*/
volatile bool GLO = 0;
// Die Enables müssen Tristate geschaltet werden, damit der GRBL-Arduino Zugriff auf das Interface hat
boolean outputmode = false;                   // Merker, ob Outputs für Enable usw. aktiviert sind
unsigned long outputmodeTimer = 0;

ISR(WDT_vect) {
	if(GLO == 0){
		digitalWrite(GREENLED,1);
		// GLO=1;
	// } else {
		// digitalWrite(GREENLED,0);
		// GLO=0;
	}
	outputmode = false;
  // if(stringComplete == false) {
    // stringComplete = true;
  // }
}
/******************************************************/

int temp;                                     // Helfer für abs()-Wandlung, die hat plötzlich ohne nicht mehr funktioniert(??)

void loop()
{
	wdt_reset();                                // zyklisch Watchdog reset auslösen. Sollte immer schnell genug sein

  if(millis() > ledblinker + 1000) {
	digitalWrite(LED, !digitalRead(LED));
  if(blinkGreen == true) { 	digitalWrite(GREENLED, !digitalRead(GREENLED));}
  if(blinkGreen == false) { 	digitalWrite(GREENLED, 0);}
	ledblinker = millis();
  }
//  if(digitalRead(SCHALTER) == 0)              // hier den SCHALTER anschließen
  if (1==1)
  {                                           // Steps nur ausgeben, wenn SCHALTER per SCHALTER auf LOW gezogen ist
    if(outputmode == false)                   // Richtung der Pins nur einmalig umschalten, wenn der Schalter betätigt wird, nicht bei jedem Durchlauf
    {
      outputmodeTimer = millis();
      // vorher die drei Enable-Ausgänge aktivieren um das Interface auf Joystick umzuschalten...
      pinMode(EnableX, OUTPUT);
      pinMode(EnableY, OUTPUT);
      pinMode(EnableZ, OUTPUT);
      digitalWrite(EnableX,0);					// die Enables sind active low
      digitalWrite(EnableY,0);
      digitalWrite(EnableZ,0);
      // ...und die Ausgangspins für Step und Dir auf OUTPUT schalten
      pinMode(stepX,OUTPUT);                    // enable step output (X-Achse)  // JA, NOTWENDIG !!!!
      pinMode(dirX,OUTPUT);                     // enable dir output (X-Achse)
      pinMode(stepY,OUTPUT);                    // enable step output (Y-Achse)
      pinMode(dirY,OUTPUT);                     // enable dir output (Y-Achse)
      pinMode(stepZ,OUTPUT);                    // enable step output (Z-Achse)
      pinMode(dirZ,OUTPUT);                     // enable dir output (Z-Achse)
      outputmode = true;
	  // digitalWrite(GREENLED,1);
    // digitalWrite(LED,1);
    }

    // ------------------------------------------------------------------------------------------------------------
    /// Alle drei Joysticks auf einen Rutsch auslesen
    joystickX = analogRead(A0);
    joystickY = analogRead(A1);   	// 512 nur fuer debug
    joystickZ = analogRead(A2);  	// 512 nur fuer debug
    // ------------------------------------------------------------------------------------------------------------
    if(lockX == false)  {
      temp = (joystickX - 512);
      temp = abs(temp);
      if((temp) > MIDDLE_INTERVAL)  {
        speedX = map((long)joystickX, (long)0, (long)1023, (long)(-((long)hugeness-(long)1)), ((long)hugeness-(long)1));                // dann skalieren
        if((speedX * directionX) < 0) digitalWrite(dirX, 0); else digitalWrite(dirX, 1);// DIR pin setzen, abhängig von directionX
        if((speedX * directionX) < 0) Xdir = -1; else Xdir = 1;
        speedX = abs(speedX);
        // PORTB |= (1 << PB5);                      // LED-Pin setzen
        sollspeedX = ((1.0 / (1 + speedX)) * (startwert + (((hugeness-1) - (speedX)) * stepsize)) + startwert);
        if(sollspeedX < oldspeedX) {              // mehr Gas
          speedtempX -= map((long)speedtempX, (long)startwert, (long)endwert, (long)1, (long)maxincspeedStep);
          if(sollspeedX >= speedtempX) {
            oldspeedX = sollspeedX;               // letzte Stepgeschwindigkeit merken
          }
        }
        if(sollspeedX > oldspeedX) {              // mehr Gas
          speedtempX += map((long)speedtempX, (long)startwert, (long)endwert, (long)1, (long)maxdecspeedStep);
          if(sollspeedX <= speedtempX) {
            oldspeedX = sollspeedX;               // letzte Stepgeschwindigkeit merken
          }
        }
        // PORTB &= ~(1 << PB5);                     // LED-Pin loeschen
      } else
      {                                           // Joystick ist in Mittelstellung -> keine Steps
        oldspeedX = endwert;                      // Einstieg mit langsamster Stepgeschwindigkeit -> sehr großer Wert
        speedtempX = 0;                           // Wert für Timer
        speedX = 0;
      }
    }
    // ------------------------------------------------------------------------------------------------------------
    if(lockY == false)  {
      temp = (joystickY - 512);
      temp = abs(temp);
      if((temp) > MIDDLE_INTERVAL)  {
        speedY = map((long)joystickY, (long)0, (long)1023, (long)(-((long)hugeness-(long)1)), ((long)hugeness-(long)1));                // dann skalieren
        if((speedY * directionY) < 0) digitalWrite(dirY, 0); else digitalWrite(dirY, 1);// DIR pin setzen, abhängig von directionX
        if((speedY * directionY) < 0) Ydir = 1; else Ydir = -1;
        speedY = abs(speedY);
        sollspeedY = ((1.0 / (1 + speedY)) * (startwert + (((hugeness-1) - (speedY)) * stepsize)) + startwert);
        if(sollspeedY < oldspeedY) {              // mehr Gas
          speedtempY -= map((long)speedtempY, (long)startwert, (long)endwert, (long)1, (long)maxincspeedStep);
          if(sollspeedY >= speedtempY) {
            oldspeedY = sollspeedY;               // letzte Stepgeschwindigkeit merken
          }
        }
        if(sollspeedY > oldspeedY) {              // mehr Gas
          speedtempY += map((long)speedtempY, (long)startwert, (long)endwert, (long)1, (long)maxdecspeedStep);
          if(sollspeedY <= speedtempY) {
            oldspeedY = sollspeedY;               // letzte Stepgeschwindigkeit merken
          }
        }
      } else
      {                                           // Joystick ist in Mittelstellung -> keine Steps
        oldspeedY = endwert;                      // Einstieg mit langsamster Stepgeschwindigkeit -> sehr großer Wert
        speedtempY = 0;                           // Wert für Timer
        speedY = 0;
      }
    }
    // ------------------------------------------------------------------------------------------------------------
    temp = (joystickZ - 512);
    temp = abs(temp);
    if((temp) > MIDDLE_INTERVAL)  {
      speedZ = map((long)joystickZ, (long)0, (long)1023, (long)(-((long)hugeness-(long)1)), ((long)hugeness-(long)1));                // dann skalieren
      if((speedZ * directionZ) < 0) digitalWrite(dirZ, 0); else digitalWrite(dirZ, 1);// DIR pin setzen, abhängig von directionX
      if((speedZ * directionZ) < 0) Zdir = 1; else Zdir = -1;
      speedZ = abs(speedZ);
      sollspeedZ = ((1.0 / (1 + speedZ)) * (startwert + (((hugeness-1) - (speedZ)) * stepsize)) + startwert);
      if(sollspeedZ < oldspeedZ) {              // mehr Gas
        speedtempZ -= map((long)speedtempZ, (long)startwert, (long)endwert, (long)1, (long)maxincspeedStep);
        if(sollspeedZ >= speedtempZ) {
          oldspeedZ = sollspeedZ;               // letzte Stepgeschwindigkeit merken
        }
      }
      if(sollspeedZ > oldspeedZ) {              // mehr Gas
        speedtempZ += map((long)speedtempZ, (long)startwert, (long)endwert, (long)1, (long)maxdecspeedStep);
        if(sollspeedZ <= speedtempZ) {
          oldspeedZ = sollspeedZ;               // letzte Stepgeschwindigkeit merken
        }
      }
    } else
    {                                           // Joystick ist in Mittelstellung -> keine Steps
      oldspeedZ = endwert;                      // Einstieg mit langsamster Stepgeschwindigkeit -> sehr großer Wert
      speedtempZ = 0;                           // Wert für Timer
      speedZ = 0;
    }
// ------------------------------------------------------------------------------------------------------------
// Grüne LED im Taster der Joystick-Box wird eingeschaltet, solange Step-Impulse erzeugt werden
    // if ((speedX != 0) || (speedY != 0) || (speedZ != 0)) {digitalWrite(GREENLED,1);} else {digitalWrite(GREENLED,0);}
// ------------------------------------------------------------------------------------------------------------

// Bedienung einer Weg-Anzeige und deren Reset mit TASTER
    if(digitalRead(TASTER) == 0)  {
      if (tasterLatch == 0) {
        tasterTime = millis();
        tasterLatch = 1;
      }
      if(millis() > tasterTime + 3000 )  {  // Taster 3 sec gedrückt -> Reset Zähler
        wegX = 0;
        wegY = 0;
        wegZ = 0;
      }
      if((blinkGreen == true) && ((lockX == true) || (lockY == true)) && (manuellMode == false))  {     // manueller Modus ist aktiv und soll abgeschaltet werden
        blinkGreen = false;
        lockX = false;
        lockY = false;
        digitalWrite(GREENLED, 1);
      }
      if((speedX != 0) && (tasterLatch == 1)) {                     // X-Achse soll freigegeben, Y-Achse gesperrt werden
        blinkGreen = true;
        lockX = false;
        lockY = true;
        manuellMode = true;
      }
      if((speedY != 0) && (tasterLatch == 1)) {                     // Y-Achse soll freigegeben, X-Achse gesperrt werden
        blinkGreen = true;
        lockX = true;
        lockY = false;
        manuellMode = true;
      }
    }
    else {
        tasterLatch = 0;
        manuellMode = false;
        tasterTime = millis();
    }
    if((debugprint == 0) && (wegvalues == 1)) { // während Debug-Ausgaben keine Weg-Werte ausgeben
      if(wegXsave != wegX) {
        Serial.print("#41M");
        wegXmm = wegX * 0.001875;
        dtostrf(wegXmm,8,4,wegBuffer);
        Serial.print(wegBuffer);
        Serial.print('<');
        wegXsave = wegX;
      }
      // xxxx
      // wegY = 0;
      // wegZ = 0;
      if(wegYsave != wegY) {
        Serial.print("#42M");
        wegYmm = wegY * 0.001875;
        dtostrf(wegYmm,8,4,wegBuffer);
        Serial.print(wegBuffer);
        Serial.print('<');
        wegYsave = wegY;
      }
      if(wegZsave != wegZ) {
        Serial.print("#43M");
        wegZmm = wegZ * 0.001875;
        dtostrf(wegZmm,8,4,wegBuffer);
        Serial.print(wegBuffer);
        Serial.print('<');
        wegZsave = wegZ;
      }
    }
    if(debugprint == 1) {         // debugprint kann über die serielle SS gesetzt werden
      Serial.print(joystickX); Serial.print("\t"); Serial.print(speedX); Serial.print("\t"); Serial.print(speedtempX); Serial.print("\t");
      Serial.print(joystickY); Serial.print("\t"); Serial.print(speedY); Serial.print("\t"); Serial.print(speedtempY); Serial.print("\t");
      Serial.print(joystickZ); Serial.print("\t"); Serial.print(speedZ); Serial.print("\t"); Serial.println(speedtempZ);
    }
  }                               // Schalter EIN hier zu Ende
  else
  {                               // Schalter AUS
    // Richtung der Pins nur einmalig umschalten, wenn der Schalter betätigt wird, nicht bei jedem Durchlauf
	if((outputmode == true) && ((outputmodeTimer + 500) < millis()))	{  // 500 millisec gegen Prellen beim Einschalten
      // die drei Enable-Ausgänge wieder deaktivieren um das Interface auf GRBL umzuschalten, wenn der SCHALTER-Schalter aus ist...
      // ...und die Ausgangspins auf INPUT (= HiZ) schalten
      pinMode(stepX,INPUT);                    	// disable step und dir output (X-Achse)
      pinMode(dirX,INPUT);
      pinMode(stepY,INPUT);                    	// disable step und dir output (Y-Achse)
      pinMode(dirY,INPUT);
      pinMode(stepZ,INPUT);                    	// disable step und dir output (Z-Achse)
      pinMode(dirZ,INPUT);
      digitalWrite(EnableX,1);                  // Enables deaktivieren (sind low activ)
      digitalWrite(EnableY,1);
      digitalWrite(EnableZ,1);
      pinMode(EnableX, INPUT);                  // Enable-Pins wieder in HiZ, damit der GRBL-Arduino Zugriff auf das Interface bekommt
      pinMode(EnableY, INPUT);
      pinMode(EnableZ, INPUT);
      outputmode = false;
      // digitalWrite(GREENLED,0);
	}
// ------------------------------------------------------------------------------------------------------------
  }

// Parameter verändern oder im EEPROM speichern oder dort holen
  if (stringComplete == true) {
    Serial.println(inputString);                                    // Echo eingegebenen String

    if((inputString.substring(0, 1) == "?") || (inputString.substring(0, 4) == "help")) { // Hilfe ausgeben
        Serial.print("Version "); Serial.print(MajorVersion); Serial.print("."); Serial.print(MinorVersion); Serial.print(", "); Serial.print(__DATE__); Serial.println(", Harald Sattler");
		// Die nachfolgende Routine liest die im Programmspeicher abgelegten Texte aus und gibt sie seriell aus
        for(i=0; i<15; i++)
        {
          memcpy_P(&pHilfeZeiger, &array[i], sizeof(PGM_P));
          strcpy_P(HilfeBuf, pHilfeZeiger);
          Serial.println(HilfeBuf);
        }
        // Die nachfolgend angegebenen Hilfetexte sind jetzt im Programmspeicher abgelegt und sparen so RAM
        /*        Serial.println("\nHilfe");
        Serial.println("Folgende Befehle werden verstanden (Gross/Kleinschreibung ist zu beachten!):");
        Serial.println("? / help     -> diese Hilfe");
        Serial.println("$$           -> listet Parameter");
        Serial.println("startwert=xx -> neuer unterer Wert (schnelles Stepping) fuer SpeedTable (*)");
        Serial.println("endwert=xxx  -> neuer oberer Wert (langsames Stepping) fuer SpeedTable (*)");
        Serial.println("stepint=xxx  -> neue Anzahl diskreter StepIntervalle (*)");
        Serial.println("dirX=x       -> Laufrichtung bei positiven Joystick X Werten. x=[-1|1] (*)");
        Serial.println("dirY=x       -> Laufrichtung bei positiven Joystick Y Werten. x=[-1|1] (*)");
        Serial.println("dirZ=x       -> Laufrichtung bei positiven Joystick Z Werten. x=[-1|1] (*)");
        Serial.println("save         -> Parameter mit(*) werden ins EEPROM gesichert");
        Serial.println("recall       -> Parameter mit(*) werden aus dem EEPROM geladen");
        Serial.println("defaults     -> Parameter werden mit Defaults vorbelegt");
        Serial.println("start        -> Debugausgaben starten");
        Serial.println("stop         -> Debugausgaben anhalten\n");
*/
        }
    if(inputString.substring(0, 2) == "$$") {                        // vorhandene Werte abfragen/ausgeben
        Serial.print("StepDir X: "); Serial.print(directionX); Serial.print("\tStepDir Y: "); Serial.print(directionY); Serial.print("\tStepDir Z: "); Serial.println(directionZ);
        Serial.print("StartWert: "); Serial.print(startwert); Serial.print("\tEndWert: "); Serial.println(endwert);
        Serial.print("StepIntervalle: "); Serial.print(hugeness); Serial.print("\tStepSize: "); Serial.println(stepsize);
        unsigned int ttt = ((1.0 / (1+ MIDDLE_INTERVAL)) * (startwert + (((hugeness-1) - (MIDDLE_INTERVAL)) * stepsize)) + startwert);
        Serial.print("SpeedValue @MIDDLE_INTERVAL ("); Serial.print(MIDDLE_INTERVAL); Serial.print("): "); Serial.println(ttt);
        Serial.print("SpeedValues: ");
        for(unsigned int j=0; j<hugeness; j++) {
              // int ttt = (((float)1/(hugeness-((hugeness-1) - i))) * (startwert + (((hugeness-1) - i) * stepsize)) + startwert);
              // int ttt = (((10000 / (hugeness-((hugeness-1) - i))) / 10000) * (startwert + (((hugeness-1) - i) * stepsize)) + startwert);
              // int ttt = (((10000L/(1+i)) * (startwert+(((hugeness-1)-i)*stepsize)))/10000)+startwert;
              ttt = ((1.0 / (1+j)) * (startwert + (((hugeness-1) - (j)) * stepsize)) + startwert);
              Serial.print(ttt); Serial.print(",");}
        Serial.println("");
        }
    if(inputString.substring(0, 9) == "startwert") {                // StartWert des Stepsize-Array ändern
        int equalSign = inputString.indexOf('=');
        inputString.remove(0,equalSign+1);
        startwert = inputString.toInt();
        Serial.print("StartWert: "); Serial.print(startwert); Serial.print("\tendwert: "); Serial.println(endwert);
        calcStepsize();                                                // setup arrays for stepsize
      }
    if(inputString.substring(0, 7) == "endwert") {                  // EndWert des Stepsize-Array ändern
        int equalSign = inputString.indexOf('=');
        inputString.remove(0,equalSign+1);
        endwert = inputString.toInt();
        Serial.print("StartWert: "); Serial.print(startwert); Serial.print("\tEndWert: "); Serial.println(endwert);
        calcStepsize();                                                // setup arrays for stepsize
      }
    if(inputString.substring(0, 7) == "stepint") {                  // EndWert des Stepsize-Array ändern
        int equalSign = inputString.indexOf('=');
        inputString.remove(0,equalSign+1);
        hugeness = inputString.toInt();
        Serial.print("StepIntervalle: "); Serial.println(hugeness);
        calcStepsize();                                                // setup arrays for stepsize
      }
    if(inputString.substring(0, 4) == "dirX") {                     // directionX ändern
        int equalSign = inputString.indexOf('=');
        inputString.remove(0,equalSign+1);
        directionX = inputString.toInt();
        Serial.print("X Richtung ");
        if(directionX == 1) Serial.print("positiv");
        if(directionX == -1) Serial.print("negativ");
      }
    if(inputString.substring(0, 4) == "dirY") {                     // directionY ändern
        int equalSign = inputString.indexOf('=');
        inputString.remove(0,equalSign+1);
        directionY = inputString.toInt();
        Serial.print("Y Richtung ");
        if(directionY == 1) Serial.print("positiv");
        if(directionY == -1) Serial.print("negativ");
      }
    if(inputString.substring(0, 4) == "dirZ") {                     // directionZ ändern
        int equalSign = inputString.indexOf('=');
        inputString.remove(0,equalSign+1);
        directionZ = inputString.toInt();
        Serial.print("Z Richtung ");
        if(directionZ == 1) Serial.print("positiv");
        if(directionZ == -1) Serial.print("negativ");
      }
    if(inputString.substring(0, 4) == "save") {                     // Parameter ins EEPROM schreiben
        Serial.println("Params -> EEPROM");
        eeAdr = 0;
        eepromWriteInt(eeAdr, startwert);
        eeAdr += 2;
        eepromWriteInt(eeAdr, endwert);
        eeAdr += 2;
        eepromWriteInt(eeAdr, directionX);
        eeAdr += 2;
        eepromWriteInt(eeAdr, directionY);
        eeAdr += 2;
        eepromWriteInt(eeAdr, directionZ);
        eeAdr += 2;
        eepromWriteInt(eeAdr, hugeness);
      }
    if(inputString.substring(0, 6) == "recall") {                   // Parameter aus EEPROM laden
        Serial.println("Params aus EEPROM laden");
        eeAdr = 0;
        startwert = eepromReadInt(eeAdr);
        eeAdr += 2;
        endwert = eepromReadInt(eeAdr);
        eeAdr += 2;
        directionX = eepromReadInt(eeAdr);
        eeAdr += 2;
        directionY = eepromReadInt(eeAdr);
        eeAdr += 2;
        directionZ = eepromReadInt(eeAdr);
        eeAdr += 2;
        hugeness = eepromReadInt(eeAdr);

        Serial.print("Startwert: "); Serial.print(startwert); Serial.print("\tEndwert: "); Serial.println(endwert);
        Serial.print("Diskrete StepIntervalle: "); Serial.print(hugeness); Serial.print("\tStepsize: "); Serial.println(stepsize);
        Serial.print("StepDir X: "); Serial.print(directionX); Serial.print("\tStepDir Y: "); Serial.print(directionY); Serial.print("\tStepDir Z: "); Serial.println(directionZ);
      }
    if(inputString.substring(0, 8) == "defaults") {                 // Default Parameter laden
        Serial.println("Default Parameter laden");
        startwert = 12;                           // unterer Grenzwert für timer
        endwert = 18000;                           // oberer Grenzwert für timer
        directionX = 1;                           // Step direction positiv bei positiven Joystickwerten
        directionY = 1;
        directionZ = 1;
        hugeness = 255;
        calcStepsize();                           // setup arrays for stepsize
        Serial.print("StartWert: "); Serial.print(startwert); Serial.print("\tEndWert: "); Serial.print(endwert); Serial.print("\tStepsize: "); Serial.println(stepsize);
        Serial.print("StepDir X: "); Serial.print(directionX); Serial.print("\tStepDir Y: "); Serial.print(directionY); Serial.print("\tStepDir Z: "); Serial.println(directionZ);
      }
    if(inputString.substring(0, 5) == "start") {                        // debug Werte auf Seriell ausgeben
        Serial.print("Debug print started");
        Serial.println("");
        debugprint = 1;
        }
    if(inputString.substring(0, 4) == "stop") {                        // debug Ausgaben stoppen
        Serial.print("Debug print stopped");
        debugprint = 0;
        }
    // clear the string
    inputString = "";
    stringComplete = false;
  }
} // loop end
// ------------------------------------------------------------------------------------------------
 // Abfrage der seriellen Schnittstelle
void serialEvent() {
    digitalWrite(GREENLED,1);
  while (Serial.available()) {
    // get the new byte:
    char inChar = (char)Serial.read();
    // add it to the inputString:
    inputString += inChar;
    // if the incoming character is a newline, set a flag
    // so the main loop can do something about it:
    if (inChar == '\n') {
      stringComplete = true;
    }
  }
    // digitalWrite(GREENLED,0);
}
// ------------------------------------------------------------------------------------------------
 // Hilfsroutinen Integer in/aus EEPROM
void eepromWriteInt(int adr, int wert) {
byte low, high;
  low=wert&0xFF;
  high=(wert>>8)&0xFF;
  EEPROM.write(adr, low); // dauert 3,3ms
  EEPROM.write(adr+1, high);
  return;
} //eepromWriteInt

int eepromReadInt(int adr) {
byte low, high;
  low=EEPROM.read(adr);
  high=EEPROM.read(adr+1);
  return low + ((high << 8)&0xFF00);
} //eepromReadInt

// ------------------------------------------------------------------------------------------------
// Hilfsroutine Stepsize-Arrays berechnen
void calcStepsize() {
   stepsize = ((endwert - startwert) / (hugeness-1));
} //calcStepsize

